##############################################################################
#   FoLiA Linguistic Annotation Tool (FLAT) - Settings file
#
# These are the settings for your FLAT site.
#
#---------------------------------------------------------------------------
#               PREAMBLE (don't edit this part)
#---------------------------------------------------------------------------
from django import VERSION as DJANGOVERSION
from socket import gethostname
import os.path
from os import environ
import flat


VERSION = flat.VERSION
try:
    BASE_DIR = os.path.dirname(os.path.dirname(flat.__file__)) + "/"
except:
    BASE_DIR = os.path.dirname(os.path.dirname(flat.__path__)) + "/"
hostname = gethostname()


##############################################################################
#               DATABASE CONFIGURATION (edit me!)
##############################################################################

#Configure your database here, by default a simple sqlite database will be used
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3', # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': 'db',                      # Or path to database file if using sqlite3.
        # The following settings are not used with sqlite3:
        'USER': '',
        'PASSWORD': '',
        'HOST': '',                      # Empty for localhost through domain sockets or '127.0.0.1' for localhost through TCP.
        'PORT': '',                      # Set to empty string for default.
    }
}


##############################################################################
#               FOLIA DOCUMENT SERVER CONFIGURATION (edit me!)
##############################################################################

#This is the path to the document root directory, this is the same directory as specified when running foliadocserve.
#If the document server is running on a different system, the remote root disk will have to be mounted and the mountpoint specified here.
WORKDIR = "/path/to/document/root/"

#The path and port on which the FoLiA Document Server can be reached (these defaults suffice for a local connection)
FOLIADOCSERVE_HOST = '127.0.0.1'
FOLIADOCSERVE_PORT = 8080

# Make sure to start the document server when starting FLAT!
#   $ foliadocserve -d /path/to/document/root -p 8080


##############################################################################
#               FLAT CONFIGURATION (edit me!)
##############################################################################


# FLAT consists of several modes, the user can select a mode from the menu.
# Each enables the user to do completely different things.
# The following modes are available and enabled. Simply
# comment the ones you want to disable:
MODES = [
    ('viewer','Viewer'),
    ('editor','Annotation Editor'),
    ('structureeditor','Structure Editor'),
    ('metadata','Metadata Editor'),
]

# The viewer and editor allow for different perspectives on the data.
#  - document: view of the entire document
#  - toc: view of a named subsection of the document (a table of contents will
#  be automatically constructed)
#  - any other FoLiA XML tag corresponsing to a structural element: paged data
#  over this type of element
PERSPECTIVES = [ 'document', 'toc', 'p', 's' ]


# Which mode should be used by default when the user opens a document?
DEFAULTMODE = 'editor'

# You may have several configurations for FLAT, a configuration determines what
# features will be enabled and what defaults will be set. Enabling all features
# will often be daunting and confusing for end-users. Using the configurations,
# you can fine-tune what users will see based on what is needed for your
# project(s).

# Users choose a configuration when logging in, the following configuration
# will be selected by default:
DEFAULTCONFIGURATION = 'full'

# Allow users to register their own accounts?
ALLOWREGISTRATION = True

# Allow public anonymous uploads without any authentication?
ALLOWPUBLICUPLOAD = False

#These are the configurations, add new ones by copying the 'full' configuration and adapting it to your needs.
CONFIGURATIONS = {
'full':{
    # This is the 'full' configuration, it will enable all available functionality and is not fine-tuned for any specific task

    # The name of the mode, this is what users will see in the login screen,
    # make sure the name is indicative of your annotation project, if any
    'name': "Full Editor",

    # use the editor modes defined globally, or specify them explicitly per configuration here.
    'modes': MODES,

    # use the perspectives defined globally, or specify them explicitly per configuration here.
    'perspectives': PERSPECTIVES,


    # Allow users to upload FoLiA documents?
    'allowupload': True, #boolean

    #Automatically create namespaces for all the groups a user is in and has write permission for,
    'creategroupnamespaces': True, #boolean

    # The default annotation focus upon loading a document
    # The annotation type that has is set as annotation focus will be
    # visualised. Set to a valid FoLiA tag name (see the FoLiA documentation at
    # https://proycon.github.io/folia) or None to start without a specific focus
    'annotationfocustype': None,
    # If you set the above, you will also need to set what FoLiA set to use
    'annotationfocusset': None,

    #EXAMPLES:
    #'annotationfocustype': 'correction',
    #'annotationfocusset': 'http://raw.github.com/proycon/folia/master/setdefinitions/spellingcorrection.foliaset.xml',

    #'annotationfocustype': 'entity',
    #'annotationfocusset': 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/namedentities.foliaset.xml',


    # Automatically enable the select span button for the annotation focus when opening the editor dialog
    'autoselectspan': False,


    # List of FoLiA annotation types (xml tags) that are allowed as annotation focus, set to True to enable all.
    # Users may switch annotation focus through the menu
    # Here and everywhere else, the annotation type equals the FoLiA XML tag
    # for the annotation, specific sets may be attached to annotation types using a slash, for example:
    #     entity/https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/namedentities.foliaset.xml
    'allowedannotationfocus': True,

    #EXAMPLE:
    #'allowedannotationfocus': ('entity/https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/namedentities.foliaset.xml'),

    # List of FoLiA annotation types (xml tags) that are initially enabled in
    # the local annotation viewer (the pop-up when the user hovers over elements)
    # set to True to enable all
    'initialviewannotations': True,


    # List of FoLiA annotation types (xml tags) that are initially enabled in
    # the global annotation viewer (in the annotation box above the words)
    'initialglobviewannotations': [],

    # List of FoLiA annotation types (xml tags) that are allowed to be viewed,
    # a superset of initialviewannotations/initialglobviewannotations. Users can enable/disable each as
    # they see fit. Set to True to enable all
    'allowedviewannotations': True,

    # List of FoLiA annotation types (xml tags) that are initially enabled in the editor dialog (when users click an element for editing)
    # set to True to enable all
    'initialeditannotations': True,

    # List of FoLiA annotation types (xml tags) that are allowed in the editor
    # dialog (the user can enable/disable each as he/she sees fit), set to True
    # to enable all
    'allowededitannotations': True,

    # Allow the user to add annotation types not yet present on a certain element?
    # If set to False, users can only edit existing annotations, never add new ones
    'allowaddfields': True, #boolean


    # Allow the user to add annotation types not yet present in the document?
    'allowdeclare': True, #boolean


    # The following values correspond to edit forms,
    # They are shown as a set of option buttons in the editor dialog, between
    # which the user can choose. Each of these options enables a button (if
    # only one is selected, no buttons appear). The edit forms can be
    # enabled/disabled by the user from a menu.

    #Enable direct editing (this is the default and most basic form of editing)
    #It should be True unless you want to force other editing forms
    'editformdirect': True, #boolean

    #Enable editing as correction
    'editformcorrection': True, #boolean

    #Enable editing as alternative
    'editformalternative': True, #boolean

    #Enable editing as new, this allows for adding multiple or overlapping annotations of the same type
    'editformnew': True, #boolean


    #This defines what edit modes the user can enable/disable from the menu
    'alloweditformdirect': True, #boolean
    'alloweditformcorrection': True, #boolean
    'alloweditformalternative': True, #boolean
    'alloweditformnew': True, #boolean

    #Allow confidence values to be set/edited?
    'allowconfidence': True,

    # FoLiA set to use for corrections
    #'initialcorrectionset': 'http://raw.github.com/proycon/folia/master/setdefinitions/spellingcorrection.foliaset.xml',

    # List of 2-tuples (tag,set) that specify what annotation types and with
    # what sets to declare automatically for each document that is opened.
    # (recall that FoLiA demands all annotations to be declared and that sets
    # can be custom-made by anyone)

    #'autodeclare': [
    #    ('correction', 'http://raw.github.com/proycon/folia/master/setdefinitions/spellingcorrection.foliaset.xml')
    #    ('entity', 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/namedentities.foliaset.xml')
    #],

    # List of 2-tuples (tag,set) that specify what annotation types and with
    # what are required sets to be already present for each document that is
    # opened. This is more or less the opposite of autodeclare

    #'requiredeclaration': [
    #    ('correction', 'http://raw.github.com/proycon/folia/master/setdefinitions/spellingcorrection.foliaset.xml')
    #    ('entity', 'https://raw.githubusercontent.com/proycon/folia/master/setdefinitions/namedentities.foliaset.xml')
    #],

    #List of metadata keys (in  FoLiA's native metadata) to show in the
    # document index.
    #'metadataindex': ['status','language','title'],

    #Dictionary of metadata keys to lists of possible values, constrains the values in the metadata editor rather than offering a free-fill field. Example:
    #'metadataconstraints': {'language': ['fr','en','es'], 'status':['completed','inprogress']},

    #'searches': [{'query': 'SELECT entity WHERE annotatortype = "manual" FOR w RETURN target', 'label': "Highlight manually annotated entities", 'changeperspective': false}]
},
}



##############################################################################
# DJANGO SETTINGS THAT NEED TO BE CHANGED (so edit me!)
#############################################################################

ADMINS = ( #Change to your contact details
    ('Maarten van Gompel', 'proycon@anaproy.nl'),
)

# Make this unique, and don't share it with anybody.
# IMPORTANT!!!! GENERATE A NEW SECRET KEY !!!! The default one here is *NOT*
# secret as it's publicly disclosed in the FLAT sources!
# (Use for instance http://www.miniwebtool.com/django-secret-key-generator/)
SECRET_KEY = 'ki5^nfv01@1g7(+*#l_0fmi9h&cf^_lv6bs4j9^6mpr&(%o4zk'

DEBUG = True #Set to False for production environments!!!!



##############################################################################
# DJANGO SETTINGS THAT NEED NOT BE CHANGED (but you may if you want to, do scroll through at least)
#############################################################################

# Feel free to tweak settings here, but the defaults should be enough

MANAGERS = ADMINS

# Hosts/domain names that are valid for this site; required if DEBUG is False
# See https://docs.djangoproject.com/en/1.5/ref/settings/#allowed-hosts
ALLOWED_HOSTS = []

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'Europe/Amsterdam'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = False

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = BASE_DIR + 'usermedia/' #not used currently by FLAT

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = 'http://flat.science.ru.nl/usermedia/' #not used currently by FLAT

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = BASE_DIR + '/flat/static/'

# URL prefix for static files.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

STYLE_ROOT = BASE_DIR + '/flat/style/'
STYLE_URL = '/style/'

# Additional locations of static files
STATICFILES_DIRS = (
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    BASE_DIR + '/flat/style/',
)


# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
#    'django.contrib.staticfiles.finders.DefaultStorageFinder',
)

LOGIN_URL = "/login/" #don't change this




# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
#     'django.template.loaders.eggs.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'flat.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = 'flat.wsgi.application'

TEMPLATE_DIRS = [
    # Put strings here, like "/home/html/django_templates" or "C:/www/django/templates".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
    BASE_DIR + 'flat/templates/'
]
for mode,_ in MODES:
    if os.path.isdir(BASE_DIR + '/flat/modes/' + mode + '/templates/'):
        TEMPLATE_DIRS.append(BASE_DIR + '/flat/modes/' + mode + '/templates/')

TEMPLATE_DIRS = tuple(TEMPLATE_DIRS)

if DJANGOVERSION[0] > 1 or DJANGOVERSION[1] >=8: #Django 1.8 and above
   TEMPLATES = [{
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'DIRS': TEMPLATE_DIRS,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },]
   del TEMPLATE_DIRS
else:
    # List of callables that know how to import templates from various sources.
    TEMPLATE_LOADERS = (
        'django.template.loaders.filesystem.Loader',
        'django.template.loaders.app_directories.Loader',
    #     'django.template.loaders.eggs.Loader',
    )
    TEMPLATE_DEBUG = DEBUG


INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    # Uncomment the next line to enable the admin:
    'django.contrib.admin',
    # Uncomment the next line to enable admin documentation:
    # 'django.contrib.admindocs',
    'flat.users'
]
for mode,_ in MODES:
    INSTALLED_APPS.append('flat.modes.' + mode)
INSTALLED_APPS = tuple(INSTALLED_APPS)


SESSION_SERIALIZER = 'django.contrib.sessions.serializers.JSONSerializer'




# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
#LOGGING = {
#    'version': 1,
#    'disable_existing_loggers': False,
#    'filters': {
#        'require_debug_false': {
#            '()': 'django.utils.log.RequireDebugFalse'
#        }
#    },
#    'handlers': {
#        'mail_admins': {
#            'level': 'ERROR',
#            'filters': ['require_debug_false'],
#            'class': 'django.utils.log.AdminEmailHandler'
#        }
#    },
#    'loggers': {
#        'django.request': {
#            'handlers': ['mail_admins'],
#            'level': 'ERROR',
#            'propagate': True,
#        },
#    }
#}



###########################################################################################################
# PROTECTION AGAINST LAZINESS, GENERAL SLOPPINESS, OVERCONFIDENCE, AND DOCUMENTATION READING IMPAIRMENTS
###########################################################################################################


#Remove this when done configuring:
raise Exception("settings.py hasn't been configured yet!!")


# Now you can start FLAT as follows:

# $ export PYTHONPATH=/your/settings/path/
# $ export DJANGO_SETTINGS_MODULE=settings
# $ django-admin runserver

# But before the first time you run you will want to run the migrations and populate the
# database:

# $ export PYTHONPATH=/your/settings/path/
# $ export DJANGO_SETTINGS_MODULE=settings
# $ django-admin makemigrations
# $ django-admin migrate --run-syncdb
# $ django-admin createsuperuser

# In any case, don't forget to start the document server too:

# $ foliadocserve -d /path/to/document/root -p 8080 --git

# If you use the --git option, make sure your git identity is configured
#    git config --global user.email "you@example.com"
#    git config --global user.name "Your Name"



from django.db import models
from django.contrib.auth.models import User, Group
from django.core.exceptions import ObjectDoesNotExist

class ReadPermissions(models.Model):
    username = models.CharField(max_length=50)
    namespace = models.CharField(max_length=50)

    def __unicode__(self):
        return self.username + " may read " + self.namespace

    def __str__(self):
        return self.username + " may read " + self.namespace

    class Meta:
        verbose_name = "Legacy read permission"
        verbose_name_plural = "Legacy read permissions"

class WritePermissions(models.Model):
    username = models.CharField(max_length=50)
    namespace = models.CharField(max_length=50)

    def __unicode__(self):
        return self.username + " may write to " + self.namespace

    def __str__(self):
        return self.username + " may write to  " + self.namespace

    class Meta:
        verbose_name = "Legacy write permission"
        verbose_name_plural = "Legacy write permissions"


def hasreadpermission(username, namespace, request):
    if username == namespace or namespace == "testflat":
        return True
    elif request.user.is_superuser:
        return True
    else:
        #new group-based system
        try:
            #is the namespace a group namespace?
            groupnamespace = Group.objects.get(name=namespace.split('/')[0])
            #are we a member of it?
            if request.user.groups.filter(name=groupnamespace.name).exists():
                return True
        except ObjectDoesNotExist:
            groupread = request.user.has_perm('auth.groupread')
            if groupread:
                try:
                    #is the namespace a user namespace?
                    usernamespace = User.objects.get(username=namespace.split('/')[0])
                    #user's namespace must share a group with us
                    if usernamespace.groups.filter(name__in=(g.name for g in request.user.groups.all())).exists():
                        return True
                except ObjectDoesNotExist:
                    pass

        #old system
        try:
            perms = ReadPermissions.objects.get(username=username, namespace='ALL')
        except:
            nsparts = namespace.split('/')
            namespace = ""
            for i, nspart in enumerate(nsparts):
                if i == 0 and nspart == username: 
                    return True
                namespace = (namespace + '/' + nspart).strip('/')
                try:
                    perms = ReadPermissions.objects.get(username=username, namespace=namespace)
                    return True
                except:
                    pass
        return False

def haswritepermission(username, namespace, request):
    if username == namespace or namespace == "testflat":
        return True
    elif request.user.is_superuser:
        return True
    else:
        #new group-based system
        try:
            #is the namespace a group namespace?
            groupnamespace = Group.objects.get(name=namespace.split('/')[0])
            #are we a member of it?
            if request.user.groups.filter(name=groupnamespace.name).exists():
                return True
        except ObjectDoesNotExist:
            groupwrite = request.user.has_perm('auth.groupwrite')
            if groupwrite:
                try:
                    #is the namespace a user namespace?
                    usernamespace = User.objects.get(username=namespace.split('/')[0])
                    #user's namespace must share a group with us
                    if usernamespace.groups.filter(name__in=(g.name for g in request.user.groups.all())).exists():
                        return True
                except ObjectDoesNotExist:
                    pass

        #old system
        try:
            perms = WritePermissions.objects.get(username=username, namespace='ALL')
        except:
            nsparts = namespace.split('/')
            namespace = ""
            for i, nspart in enumerate(nsparts):
                if i == 0 and nspart == username:
                    return True
                namespace = (namespace + '/' + nspart).strip('/')
                try:
                    perms = WritePermissions.objects.get(username=username, namespace=namespace)
                    return True
                except:
                    pass
        return False



from django import VERSION

from django.contrib import admin, auth

if VERSION[0] >= 2 or VERSION[1] >= 8: #Django 1.8 and higher
    from django.conf.urls import url, include
    from django.conf.urls.static import static
elif VERSION[1] >= 6: #Django 1.6
    from django.conf.urls import patterns, url, include
    from django.conf.urls.static import static
else:
    from django.conf.urls.defaults import *

from django.conf import settings
import flat.views
admin.autodiscover()

urlpatterns = [
    # Examples:
    url(r'^$', flat.views.index, name='index'),
    url(r'^index/?$', flat.views.index, name='index'),
    url(r'^pub/?$', flat.views.pub, name='pub'),
    url(r'^index/(?P<namespace>[\w\d\-_\./@]+)/?$', flat.views.index, name='index'),
    url(r'^login/?$', flat.views.login, name='login'),
    url(r'^logout/?$', flat.views.logout, name='logout'),
    url(r'^account/', include('django.contrib.auth.urls')),
    url(r'^admin/', include(admin.site.urls)),
    url(r'^admin/doc/', include('django.contrib.admindocs.urls')),
    url(r'^register/?$', flat.views.register, name='register'),
    url(r'^download/pub/(?P<docid>[\w\d\-_\.]+)\.folia\.xml/?$', flat.views.pub_download, name='download'),
    url(r'^download/pub/(?P<docid>[\w\d\-_\.]+)/?$', flat.views.pub_download, name='download'),
    url(r'^download/(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)\.folia\.xml/?$', flat.views.download, name='download'),
    url(r'^download/(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/?$', flat.views.download, name='download'),
    url(r'^pub/upload/?$', flat.views.pub_upload, name='pub_upload'),
    url(r'^upload/?$', flat.views.upload, name='upload'),
    url(r'^filemanagement/?$', flat.views.filemanagement, name='filemanagement'),
    url(r'^addnamespace/?$', flat.views.addnamespace, name='addnamespace'),
    url(r'^pub/(?P<configuration>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/query/?$', flat.views.pub_query, name='query'), #generic query function
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/query/?$', flat.views.query, name='query'), #generic query function
    # url(r'^flat/', include('flat.foo.urls')),

]

if VERSION[0] == 1 and VERSION[1] < 8: #Django <1.8
    urlpatterns = patterns('',*urlpatterns)

if settings.DEBUG:
    if VERSION[0] > 1 or VERSION[1] >= 6: #Django 1.6
        urlpatterns += static(settings.STYLE_URL, document_root=settings.STYLE_ROOT)
    else:
        #Django 1.3
        urlpatterns += patterns('',
            (r'^style/(?P<path>.*)$', 'django.views.static.serve', {'document_root': settings.STYLE_ROOT}),
        )


for mode, _ in settings.MODES:
    urlpatterns += [ url(r'^' + mode + '/', include('flat.modes.' + mode + '.urls'))]
    if VERSION[0] == 1 and VERSION[1] < 8: #Django <1.8
        urlpatterns += patterns('', url(r'^' + mode + '/', include('flat.modes.' + mode + '.urls')))
    else:
        urlpatterns += [ url(r'^' + mode + '/', include('flat.modes.' + mode + '.urls'))]






import sys
if sys.version < '3':
    from urllib import urlencode
    from urllib2 import urlopen, Request
    from StringIO import StringIO
else:
    from urllib.parse import urlencode
    from urllib.request import urlopen, Request
    from io import BytesIO as StringIO
from django.conf import settings
import json
import requests


REQUIREFOLIADOCSERVE = '0.6'

def checkversion(version):
    """Checks foliadocserve version, returns 1 if the document is newer than the library, -1 if it is older, 0 if it is equal"""
    try:
        for refversion, responseversion in zip([int(x) for x in REQUIREFOLIADOCSERVE.split('.')], [int(x) for x in version.split('.')]):
            if responseversion > refversion:
                return 1 #response is newer than library
            elif responseversion < refversion:
                return -1 #response is older than library
        return 0 #versions are equal
    except ValueError:
        raise ValueError("Unable to parse version, invalid syntax")


def getsid(request):
    if 'HTTP_X_SESSIONID' in request.META:
        return request.META['HTTP_X_SESSIONID'] #accesses X-Sessionid header
    elif 'sid' in request.POST:
        return request.POST['sid']
    elif 'sid' in request.GET:
        return request.GET['sid']
    else:
        return 'NOSID'

def setsid(request, sid):
    request.add_header('X-sessionid', sid)

def query(request, query, parsejson=True, **extradata):
    data = {'query':query}
    for key, value in extradata.items():
        if isinstance(value, bool):
            data[key] = int(value)
        else:
            data[key] = value
    if sys.version < '3':
        #encode for python 2
        for key, value in data.items():
            if isinstance(value,unicode): #pylint: disable=undefined-variable
                data[key] = value.encode('utf-8')

    docservereq = Request("http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/query/")
    setsid(docservereq, getsid(request))
    f = urlopen(docservereq,urlencode(data).encode('utf-8')) #or opener.open()
    if sys.version < '3':
        contents = unicode(f.read(),'utf-8') #pylint: disable=undefined-variable
    else:
        contents = str(f.read(),'utf-8')
    f.close()
    if contents and contents[0] in ('{','['):
        #assume this is json
        if parsejson:
            parsed = json.loads(contents)
            if 'version' not in parsed:
                raise Exception("No version information supplied by foliadocserve, it is likely too old, please upgrade")
            elif checkversion(parsed['version']) == -1:
                raise Exception("Foliadocserve version is too old, got " + parsed['version'] + ", expected at " + REQUIREFOLIADOCSERVE + ", please upgrade")
            return parsed
        else:
            return contents
    elif contents:
        return contents
    else:
        if parsejson:
            return None
        else:
            return ""




def get( request, url, parsejson=True):
    docservereq = Request("http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/" + url) #or opener.open()
    setsid(docservereq, getsid(request))
    f = urlopen(docservereq)
    if sys.version < '3':
        contents = unicode(f.read(),'utf-8')
    else:
        contents = str(f.read(),'utf-8')
    f.close()
    if contents and contents[0] in ('{','['):
        #assume this is json
        if parsejson:
            return json.loads(contents)
        else:
            return contents
    elif contents:
        return contents
    else:
        if parsejson:
            return None
        else:
            return ""

def filemanagement(request, filemanmode, namespace, doc, **data):
    if sys.version < '3':
        #encode for python 2
        for key, value in data.items():
            if isinstance(value,unicode): #pylint: disable=undefined-variable
                data[key] = value.encode('utf-8')

    docservereq = Request("http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/" + filemanmode + "/" + namespace + "/" + doc)
    setsid(docservereq, getsid(request))
    f = urlopen(docservereq,urlencode(data).encode('utf-8')) #or opener.open()
    if sys.version < '3':
        contents = unicode(f.read(),'utf-8') #pylint: disable=undefined-variable
    else:
        contents = str(f.read(),'utf-8')
    f.close()
    if contents and contents[0] in ('{','['):
        return json.loads(contents)
    else:
        return None


def postjson( request, url, data):
    if isinstance(data, dict) or isinstance(data,list) or isinstance(data, tuple):
        data = json.dumps(data)
    docservereq = Request("http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/" + url + '/' + sid) #or opener.open()
    setsid(docservereq, getsid(request))
    docservereq.add_header('Content-Type', 'application/json')
    f = urlopen(docservereq, urlencode(data).encode('utf-8'))
    if sys.version < '3':
        contents = unicode(f.read(),'utf-8')
    else:
        contents = str(f.read(),'utf-8')
    f.close()
    if contents and contents[0] == '{':
        #assume this is json
        return json.loads(contents)
    elif contents:
        return contents
    else:
        return None

def postxml( request, url, f):
    #buf = StringIO()
    #c = pycurl.Curl()
    #url = "http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/" + url
    #c.setopt(c.URL, url.encode('utf-8'))
    #c.setopt(c.HTTPPOST, data.encode('utf-8') )
    #c.setopt(pycurl.HTTPHEADER, ["Content-Type: application/json"])
    #c.setopt(pycurl.POST, 1)
    #c.setopt(pycurl.POSTFIELDS, data.encode('utf-8'))
    #c.setopt(c.WRITEFUNCTION, buf.write)
    #c.perform()
    #code = c.getinfo(c.HTTP_CODE)
    #contents = buf.getvalue()
    #c.close()


    #req = Request("http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/" + url)
    #req.add_header('Content-Type', 'application/xml; charset=utf-8')
    #req.add_data(urlencode({'data':data.encode('utf-8')}))
    #f = urlopen(req)
    #contents = f.read()
    #f.close()

    response = requests.post("http://" + settings.FOLIADOCSERVE_HOST + ":" + str(settings.FOLIADOCSERVE_PORT) + "/" + url, data=f, headers={'Content-Type':'application/xml; charset=utf-8'})
    contents = response.text
    if contents and contents[0] == '{':
        #assume this is json
        return json.loads(contents)
    elif contents:
        return contents
    else:
        return None




from __future__ import print_function, unicode_literals, division, absolute_import
import sys
import os
import json
import datetime
import logging

from collections import defaultdict

from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required
from django.views.decorators.csrf import csrf_exempt
from django.http import HttpResponse, HttpResponseForbidden, HttpResponseRedirect, HttpRequest
import django.contrib.auth
from django.conf import settings

from pynlpl.formats import fql

import flat.comm
import flat.users
from flat.converters import get_converters, inputformatchangefunction
from flat.modes.metadata.models import MetadataIndex
import lxml.html
if sys.version < '3':
    from urllib2 import URLError, HTTPError #pylint: disable=import-error
else:
    from urllib.error import URLError, HTTPError

# Get an instance of a logger
logger = logging.getLogger(__name__)

def getcontext(request,namespace,docid, doc, mode, configuration = None):
    if configuration is None:
        if 'configuration' not in request.session:
            raise Exception("Session has been logged out")
        else:
            configuration = request.session['configuration']
    return {
            'configuration_id': configuration,
            'configuration': settings.CONFIGURATIONS[configuration],
            'configuration_json': json.dumps(settings.CONFIGURATIONS[configuration]),
            'namespace': namespace,
            'testnum': request.GET.get('testNumber',0),
            'docid': docid,
            'mode': mode,
            'modes': settings.CONFIGURATIONS[configuration]['modes'] ,
            'modes_json': json.dumps([x[0] for x in settings.CONFIGURATIONS[configuration]['modes'] ]),
            'perspectives_json': json.dumps(settings.CONFIGURATIONS[configuration]['perspectives']),
            'docdeclarations': json.dumps(doc['declarations']) if 'declarations' in doc else "{}",
            'setdefinitions': json.dumps(doc['setdefinitions']) if 'setdefinitions' in doc else "{}",
            'metadata': json.dumps(doc['metadata']) if 'metadata' in doc else "{}",
            'toc': json.dumps(doc['toc']) if 'toc' in doc else "[]",
            'slices': json.dumps(doc['slices']) if 'slices' in doc else "{}",
            'slicesize': json.dumps(doc['slicesize']) if 'slicesize' in doc else "{}",
            'rtl': True if 'rtl' in doc and doc['rtl'] else False,
            'loggedin': request.user.is_authenticated() if request.user else False,
            'isadmin': request.user.is_staff if request.user else False,
            'version': settings.VERSION,
            'username': request.user.username if namespace != "pub" else "anonymous",
            'waitmessage': "Loading document on server and initialising web-environment...",
    }

def validatenamespace(namespace):
    return namespace.replace('..','').replace('"','').replace(' ','_').replace(';','').replace('&','').strip('/')

def getdocumentselector(query):
    if query.startswith("USE "):
        end = query[4:].index(' ') + 4
        if end >= 0:
            try:
                namespace,docid = query[4:end].rsplit("/",1)
            except:
                raise fql.SyntaxError("USE statement takes namespace/docid pair")
            return (validatenamespace(namespace),docid), query[end+1:]
        else:
            try:
                namespace,docid = query[4:end].rsplit("/",1)
            except:
                raise fql.SyntaxError("USE statement takes namespace/docid pair")
            return (validatenamespace(namespace),docid), ""
    return None, query

def getbody(html):
    doc = lxml.html.fromstring(html)
    return doc.xpath("//body/p")[0].text_content()

def docserveerror(e, d=None):
    if d is None: d={}
    if isinstance(e, HTTPError):
        body = getbody(e.read())
        d['fatalerror'] =  "<strong>Fatal Error:</strong> The document server returned an error<pre style=\"font-weight: bold\">" + str(e) + "</pre><pre>" + body +"</pre>"
        d['fatalerror_text'] = body
    elif isinstance(e, URLError):
        d['fatalerror'] =  "<strong>Fatal Error:</strong> Could not connect to document server!"
        d['fatalerror_text'] =  "Could not connect to document server!"
    elif isinstance(e, str) :
        if sys.version < '3':
            d['fatalerror'] =   e.decode('utf-8') #pylint: disable=undefined-variable
            d['fatalerror_text'] = e.decode('utf-8') #pylint: disable=undefined-variable
        else:
            d['fatalerror'] =  e
            d['fatalerror_text'] = e
    elif sys.version < '3' and isinstance(e, unicode): #pylint: disable=undefined-variable
        d['fatalerror'] =  e
        d['fatalerror_text'] = e
    elif isinstance(e, Exception):
        # we don't handle other exceptions, raise!
        raise e
    return d


def initdoc(request, namespace, docid, mode, template, context=None, configuration=None):
    """Initialise a document (not invoked directly)"""
    perspective = request.GET.get('perspective','document')
    if context is None: context = {}
    if 'configuration' in request.session:
        configuration = request.session['configuration']
    elif configuration is None:
        return fatalerror(request, "No configuration specified")
    if configuration not in settings.CONFIGURATIONS:
        return fatalerror(request, "Specified configuration does not exist")
    flatargs = {
        'setdefinitions': True,
        'declarations': True,
        'metadata': True,
        'toc': True,
        'slices': request.GET.get('slices',settings.CONFIGURATIONS[configuration].get('slices','p:25,s:100')), #overriden either by configuration or by user
        'customslicesize': 0, #disabled for initial probe
        'textclasses': True,
    }
    error = False
    try:
        doc = flat.comm.query(request, "USE " + namespace + "/" + docid + " PROBE", **flatargs) #retrieves only the meta information, not document content
        context.update(getcontext(request,namespace,docid, doc, mode, configuration))
    except Exception as e:
        context.update(docserveerror(e))
        error = True

    if not error:
        dorequiredeclaration = 'requiredeclaration' in settings.CONFIGURATIONS[configuration] and settings.CONFIGURATIONS[configuration]['requiredeclaration']
        if dorequiredeclaration:
            if not 'declarations' in doc:
                return fatalerror(request, "Refusing to load document, missing expected declarations, none declared")
            declarations = doc['declarations']
            for annotationtype, annotationset in settings.CONFIGURATIONS[configuration]['requiredeclaration']:
                found = False
                for d in declarations:
                    if annotationtype == d['annotationtype'] and (not annotationset or annotationset == d['set']):
                        found = True
                        break
                if not found:
                    if annotationset:
                        return fatalerror(request, "Refusing to load document, missing expected declaration for annotation type " + annotationtype + "/" + annotationset)
                    else:
                        return fatalerror(request, "Refusing to load document, missing expected declaration for annotation type " + annotationtype)

        dometadataindex = 'metadataindex' in settings.CONFIGURATIONS[configuration] and settings.CONFIGURATIONS[configuration]['metadataindex']
        if dometadataindex:
            metadata = json.loads(context['metadata'])
            for metakey in settings.CONFIGURATIONS[configuration]['metadataindex']:
                if metakey in metadata:
                    MetadataIndex.objects.update_or_create(namespace=namespace,docid=docid, key=metakey,defaults={'value':metadata[metakey]})
    response = render(request, template, context)
    if 'fatalerror' in context:
        response.status_code = 500
    return response

def query_helper(request,namespace, docid, configuration=None):
    """Does the actual query, called by query() or pub_query(), not directly"""
    flatargs = {
        'customslicesize': request.POST.get('customslicesize',settings.CONFIGURATIONS[configuration].get('customslicesize','50')), #for pagination of search results
    }
    #stupid compatibility stuff
    if sys.version < '3':
        if hasattr(request, 'body'):
            data = json.loads(unicode(request.body,'utf-8')) #pylint: disable=undefined-variable
        else: #older django
            data = json.loads(unicode(request.raw_post_data,'utf-8')) #pylint: disable=undefined-variable
    else:
        if hasattr(request, 'body'):
            data = json.loads(str(request.body,'utf-8'))
        else: #older django
            data = json.loads(str(request.raw_post_data,'utf-8'))

    if not data['queries']:
        return HttpResponseForbidden("No queries to run")

    for query in data['queries']:
        #get document selector and check it doesn't violate the namespace
        docselector, query = getdocumentselector(query)
        if not docselector:
            return HttpResponseForbidden("Query does not start with a valid document selector (USE keyword)!")
        elif docselector[0] != namespace:
            return HttpResponseForbidden("Query would affect a different namespace than your current one, forbidden!")

        if query != "GET" and query[:4] != "CQL " and query[:4] != "META":
            #parse query on this end to catch syntax errors prior to sending, should be fast enough anyway
            try:
                query = fql.Query(query)
            except fql.SyntaxError as e:
                return HttpResponseForbidden("FQL Syntax Error: " + str(e))
            needwritepermission = query.declarations or query.action and query.action.action != "SELECT"
        else:
            needwritepermission = False

    if needwritepermission and configuration != "pub" and not flat.users.models.haswritepermission(request.user.username, namespace, request):
        return HttpResponseForbidden("Permission denied, no write access")

    query = "\n".join(data['queries']) #throw all queries on a big pile to transmit
    try:
        d = flat.comm.query(request, query,**flatargs)
    except Exception as e:
        if sys.version < '3':
            errmsg = docserveerror(e)['fatalerror_text']
            return HttpResponseForbidden("FoLiA Document Server error: ".encode('utf-8') + errmsg.encode('utf-8'))
        else:
            return HttpResponseForbidden("FoLiA Document Server error: " + docserveerror(e)['fatalerror_text'])
    return HttpResponse(json.dumps(d).encode('utf-8'), content_type='application/json')



@login_required
def query(request,namespace, docid):
    if request.method != 'POST':
        return HttpResponseForbidden("POST method required for " + namespace + "/" + docid + "/query")

    if flat.users.models.hasreadpermission(request.user.username, namespace, request):
        return query_helper(request, namespace,docid, request.session['configuration'])
    else:
        return HttpResponseForbidden("Permission denied, no read access")

@csrf_exempt
def pub_query(request,configuration, docid):
    if request.method != 'POST':
        return HttpResponseForbidden("POST method required for pub/" + docid + "/query")

    return query_helper(request, 'pub',docid, configuration)

def login(request):
    if 'username' in request.POST and 'password' in request.POST:
        username = request.POST['username']
        password = request.POST['password']
        request.session['configuration'] = request.POST['configuration']
        user = django.contrib.auth.authenticate(username=username, password=password)
        if user is not None:
            if user.is_active:
                django.contrib.auth.login(request, user)
                # Redirect to a success page.
                if 'next' in request.POST:
                    return redirect("/" + request.POST['next'])
                elif 'next' in request.GET:
                    return redirect("/" + request.GET['next'])
                else:
                    return redirect("/")
            else:
                # Return a 'disabled account' error message
                return render(request, 'login.html', {'error': "This account is disabled","defaultconfiguration":settings.DEFAULTCONFIGURATION, "configurations":settings.CONFIGURATIONS , 'version': settings.VERSION, "allowregistration": hasattr(settings, 'ALLOWREGISTRATION') and settings.ALLOWREGISTRATION} )
        else:
            # Return an 'invalid login' error message.
            return render(request, 'login.html', {'error': "Invalid username or password","defaultconfiguration":settings.DEFAULTCONFIGURATION, "configurations":settings.CONFIGURATIONS, 'version': settings.VERSION, "allowregistration": hasattr(settings, 'ALLOWREGISTRATION') and settings.ALLOWREGISTRATION} )
    else:
        return render(request, 'login.html',{"defaultconfiguration":settings.DEFAULTCONFIGURATION, "configurations":settings.CONFIGURATIONS, "version": settings.VERSION, "allowregistration": hasattr(settings, 'ALLOWREGISTRATION') and settings.ALLOWREGISTRATION})


def logout(request):
    if 'configuration' in request.session:
        del request.session['configuration']
    django.contrib.auth.logout(request)
    return redirect("/login")


def register(request):
    if hasattr(settings, 'ALLOWREGISTRATION') and not settings.ALLOWREGISTRATION:
        return HttpResponseForbidden("Registration disabled")

    if request.method == 'POST':
        form = django.contrib.auth.forms.UserCreationForm(request.POST)
        if form.is_valid():
            new_user = form.save()
            return HttpResponseRedirect("/login/")
    else:
        form = django.contrib.auth.forms.UserCreationForm()
    return render(request, "register.html", {
        'form': form,
        'version': settings.VERSION,
    })

def fatalerror(request, e,code=404):
    assert isinstance(request, HttpRequest)
    if isinstance(e, Exception):
        response = render(request,'base.html', docserveerror(e))
    else:
        response = render(request,'base.html', {'fatalerror': str(e)})
    response.status_code = code
    return response


@login_required
def index(request, namespace=""):
    try:
        namespaces = flat.comm.get(request, '/namespaces/')
    except Exception as e:
        return fatalerror(request,e)

    if not 'configuration' in request.session:
        return logout(request)

    if not namespace:
        #check if user namespace is preset, if not, make it
        if not request.user.username in namespaces['namespaces']:
            try:
                flat.comm.get(request, "createnamespace/" + request.user.username, False)
            except Exception as e:
                return fatalerror(request,e)

        if 'creategroupnamespaces' in settings.CONFIGURATIONS[request.session['configuration']] and settings.CONFIGURATIONS[request.session['configuration']]['creategroupnamespaces'] and request.user.has_perm('groupwrite'):
            for group in request.user.groups.all():
                try:
                    flat.comm.get(request, "createnamespace/" + group.name, False)
                except Exception as e:
                    return fatalerror(request,e)

        if request.user.has_perm('auth.groupwrite') and (request.user.has_perm('auth.allowcopy') or request.user.has_perm('auth.allowdelete')):
            #create namespaces for all the users in our groups, since we may want to copy there
            for group in request.user.groups.all():
                for user in django.contrib.auth.models.User.objects.filter(groups__name=group.name):
                    try:
                        flat.comm.get(request, "createnamespace/" + user.username, False)
                    except Exception as e:
                        return fatalerror(request,e)

    readpermission = flat.users.models.hasreadpermission(request.user.username, namespace, request)
    dirs = []
    recursivedirs =  []
    subdirs =[]
    for ns in sorted(namespaces['namespaces']):
        if flat.users.models.hasreadpermission(request.user.username, ns, request):
            recursivedirs.append(ns)
            if '/' not in ns:
                dirs.append(ns)
                if not namespace:
                    subdirs.append(ns)
            if ns.startswith(namespace + '/'):
                subdirs.append(ns[len(namespace)+1:])
    dirs.sort()
    recursivedirs.sort()

    dometadataindex = 'metadataindex' in settings.CONFIGURATIONS[request.session['configuration']] and settings.CONFIGURATIONS[request.session['configuration']]['metadataindex']
    if dometadataindex:
        metadataindex = defaultdict(list)
        for m in MetadataIndex.objects.filter(namespace=namespace):
            metadataindex[m.docid].append( (m.key, m.value) )

    docs = []
    if namespace and readpermission:
        try:
            r = flat.comm.get(request, '/documents/' + namespace)
        except Exception as e:
            return fatalerror(request,e)
        for d in sorted(r['documents']):
            docid =  os.path.basename(d.replace('.folia.xml',''))
            if dometadataindex and docid in metadataindex:
                metaitems = metadataindex[docid]
            else:
                metaitems = []
            docs.append( (docid, round(r['filesize'][d] / 1024 / 1024,2) , datetime.datetime.fromtimestamp(r['timestamp'][d]).strftime("%Y-%m-%d %H:%M"), metaitems ) )


    docs.sort()

    if namespace:
        parentdir = '/'.join(namespace.split('/')[:-1])
    else:
        parentdir = ""

    return render(request, 'index.html', {'namespace': namespace,'parentdir': parentdir, 'dirs': dirs, 'recursivedirs': recursivedirs, 'subdirs': subdirs, 'docs': docs, 'defaultmode': settings.DEFAULTMODE,'loggedin': request.user.is_authenticated(), 'isadmin': request.user.is_staff, 'username': request.user.username, 'configuration': settings.CONFIGURATIONS[request.session['configuration']], 'converters': get_converters(request), 'inputformatchangefunction': inputformatchangefunction(request), 'allowcopy': request.user.has_perm('auth.allowcopy'), 'allowdelete': request.user.has_perm('auth.allowdelete'),'version': settings.VERSION})

def pub(request):
    if hasattr(settings,'ALLOWPUBLICUPLOAD') and settings.ALLOWPUBLICUPLOAD:
        return render(request, 'pub.html', {'defaultmode': settings.DEFAULTMODE,'loggedin': request.user.is_authenticated(), 'isadmin': request.user.is_staff, 'username': request.user.username,"defaultconfiguration":settings.DEFAULTCONFIGURATION, "configurations":settings.CONFIGURATIONS, 'version': settings.VERSION})
    else:
        return fatalerror(request, "Public anonymous write permission denied",403)

@login_required
def download(request, namespace, docid):
    data = flat.comm.query(request, "USE " + namespace + "/" + docid + " GET",False)
    return HttpResponse(data, content_type='text/xml')

@csrf_exempt
def pub_download(request, docid):
    data = flat.comm.query(request, "USE pub/" + docid + " GET",False)
    return HttpResponse(data, content_type='text/xml')

@login_required
def filemanagement(request):
    if request.method == 'POST':
        if 'filemanmode' in request.POST:
            if request.POST['filemanmode'] == 'delete':
                targetnamespace = None
            elif request.POST['filemanmode'] == 'copy' and 'copytarget' in request.POST and request.POST['copytarget']:
                targetnamespace = request.POST['copytarget'].replace('..','.').replace(' ','').replace('&','')
            elif request.POST['filemanmode'] == 'move' and 'movetarget' in request.POST and request.POST['movetarget']:
                targetnamespace = request.POST['movetarget'].replace('..','.').replace(' ','').replace('&','')
            else:
                return fatalerror(request, "Invalid operation",404)

            docs = []
            for key, value in request.POST.items():
                if key.startswith('docselect'):
                    docs.append(value)
            if not docs:
                return fatalerror(request, "No documents selected",404)

            if targetnamespace:
                if not flat.users.models.haswritepermission(request.user.username, targetnamespace, request):
                    return fatalerror(request, "Write permission denied for " + targetnamespace,403)

            namespace = None
            for doc in docs:
                namespace, docid = doc.rsplit('/', 1)
                if targetnamespace:
                    data = {'target': targetnamespace + '/' + docid}
                else:
                    data = {}

                if not flat.users.models.hasreadpermission(request.user.username, namespace, request):
                    return fatalerror(request, "Read permission denied for " + docid,403)
                if request.POST['filemanmode'] in ('delete','move') and not flat.users.models.haswritepermission(request.user.username, namespace, request):
                    return fatalerror(request, "Write permission denied for " + docid,403)

                try:
                    flat.comm.filemanagement(request,request.POST['filemanmode'], namespace, docid, **data)
                except Exception as e:
                    return fatalerror(request,e)

            if targetnamespace:
                return HttpResponseRedirect("/index/" + targetnamespace )
            elif namespace:
                return HttpResponseRedirect("/index/" + namespace )
            else:
                return HttpResponseRedirect("/")

        else:
            return fatalerror(request, "No mode specified",403)
    else:
        return fatalerror(request, "Method denied",403)


def upload_helper(request, namespace, configuration=None, mode=None):
    if not mode:
        mode = settings.DEFAULTMODE
    if not configuration:
        configuration = settings.DEFAULTCONFIGURATION
    if mode not in dict(settings.MODES):
        return fatalerror(request, "Invalid mode",403)
    if configuration not in settings.CONFIGURATIONS:
        return fatalerror(request, "Invalid configuration",403)

    if 'inputformat' in request.POST and request.POST['inputformat'] != 'folia':
        #we need to convert the input
        converter = None
        for cv in get_converters(request):
            if request.POST['inputformat'] == cv.id:
                converter = cv
        if not converter:
            return fatalerror(request, "Converter not found or specified",404)

        try:
            parameters = converter.parse_parameters(request, 'parameters')
        except:
            return fatalerror(request, "Invalid syntax for conversion parameters",403)

        if 'TMPDIR' in os.environ:
            tmpdir = os.environ['TMPDIR']
        else:
            tmpdir = '/tmp'

        tmpinfile = os.path.join(tmpdir, request.FILES['file'].name)
        with open(tmpinfile,'wb') as f_tmp:
            for chunk in request.FILES['file'].chunks():
                f_tmp.write(chunk)

        tmpoutfile = converter.get_output_name(os.path.join(tmpdir, request.FILES['file'].name))
        success, msg = converter.convert(tmpinfile, tmpoutfile, **parameters)
        if not success:
            return fatalerror(request, "Input conversion failed: " + msg,403)

        #removing temporary input file
        os.unlink(tmpinfile)

        #reading temporary output file
        f_folia =  open(tmpoutfile, 'rb')
    else:
        f_folia = request.FILES['file']

    #if sys.version < '3':
    #    data = unicode(request.FILES['file'].read(),'utf-8') #pylint: disable=undefined-variable
    #else:
    #    data = str(request.FILES['file'].read(),'utf-8')
    failed = False
    try:
        response = flat.comm.postxml(request,"upload/" + namespace , f_folia)
    except Exception as e:
        failed = True
        response = fatalerror(request,e)

    if 'inputformat' in request.POST and request.POST['inputformat'] != 'folia':
        f_folia.close()
        #removing temporary output file after conversion
        os.unlink(tmpoutfile)

    if failed:
        return response
    elif 'error' in response and response['error']:
        return fatalerror(request, response['error'],403)
    else:
        docid = response['docid']
        logger.info("Succesfully uploaded " + docid + "(namespace=" + namespace+")")
        if namespace == "pub":
            return HttpResponseRedirect("/" + mode + "/pub/" + configuration + "/" + docid  )
        else:
            return HttpResponseRedirect("/" + mode + "/" + namespace + "/" + docid  )

@login_required
def upload(request):
    if request.method == 'POST':
        namespace = request.POST['namespace'].replace('..','.').replace(' ','').replace('&','')
        if flat.users.models.haswritepermission(request.user.username, namespace, request) and 'file' in request.FILES:
            return upload_helper(request, namespace)
        else:
            return fatalerror(request, "Write permission denied",403)
    else:
        return fatalerror(request, "Method denied",403)


@csrf_exempt
def pub_upload(request):
    if request.method == 'POST':
        if hasattr(settings,'ALLOWPUBLICUPLOAD') and settings.ALLOWPUBLICUPLOAD:
            if 'configuration' in request.POST:
                configuration = request.POST['configuration']
            else:
                configuration = settings.DEFAULTCONFIGURATION
            if 'mode' in request.POST:
                mode = request.POST['mode']
            else:
                mode = settings.DEFAULTMODE
            logger.info("Receiving public upload for configuration " + configuration + ", mode " + mode)
            return upload_helper(request, 'pub', configuration, mode)
        else:
            return fatalerror(request, "Public anonymous write permission denied",403)
    else:
        return fatalerror(request, "Method denied",403)

@login_required
def addnamespace(request):
    if request.method == 'POST':
        namespace = request.POST['namespace'].replace('/','').replace('..','.').replace(' ','').replace('&','')
        newdirectory = request.POST['newdirectory'].replace('/','').replace('..','.').replace(' ','').replace('&','')
        if flat.users.models.haswritepermission(request.user.username, namespace, request):
            try:
                response = flat.comm.get(request,"createnamespace/" + namespace + "/" + newdirectory)
            except Exception as e:
                return fatalerror(request,e)
            if 'error' in response and response['error']:
                return fatalerror(request, response['error'],403)
            elif namespace:
                return HttpResponseRedirect("/index/" + namespace + '/' + newdirectory  )
            else:
                return HttpResponseRedirect("/index/" + newdirectory  )
        else:
            return fatalerror(request, "Permission denied",403)
    else:
        return fatalerror(request, "Permission denied",403)



#from django.conf.urls import patterns, include, url
from django import VERSION
if VERSION[0] > 1 or VERSION[1] >= 8: #Django 1.8 and higher
    from django.conf.urls import url, include
elif VERSION[1] >= 6: #Django 1.6
    from django.conf.urls import patterns, url, include
else:
    from django.conf.urls.defaults import *

import flat.modes.structureeditor.views

urlpatterns = [
    # Examples:
    url(r'^(?P<namespace>[\w\-_\.@]+)/(?P<docid>[\w\-\.]+)/?$', flat.modes.structureeditor.views.view, name='view'),
    #url(r'^(?P<namespace>[\w\-\.]+)/(?P<docid>[\w\-\.]+)/annotate/?$', 'flat.modes.structureeditor.views.annotate', name='annotate'),
]

if VERSION[0] == 1 and VERSION[1] < 8: #Django <1.8
    urlpatterns = patterns('',*urlpatterns)



from __future__ import print_function, unicode_literals, division, absolute_import
from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse,HttpResponseForbidden
import sys
if sys.version < '3':
    from urllib2 import URLError
else:
    from urllib.error import URLError
from django.conf import settings
import flat.comm
import flat.users
from flat.views import initdoc, fatalerror
import json


@login_required
def view(request, namespace, docid):
    """The initial view, does not provide the document content yet"""
    if flat.users.models.hasreadpermission(request.user.username, namespace, request):
        if 'autodeclare' in settings.CONFIGURATIONS[request.session['configuration']]:
            if flat.users.models.haswritepermission(request.user.username, namespace, request):
                for annotationtype, set in settings.CONFIGURATIONS[request.session['configuration']]['autodeclare']:
                    try:
                        r = flat.comm.query(request, "USE " + namespace + "/" + docid + " DECLARE " + annotationtype + " OF " + set)
                    except Exception as e:
                        return fatalerror(request,e)

        return initdoc(request, namespace,docid, 'metadata', 'metadata.html')
    else:
        return fatalerror(request,"Permission denied")




from django.dispatch import receiver
from django.db.models.signals import post_migrate
from django.core.exceptions import ObjectDoesNotExist

#hacky solution to ensure custom permissions are introduce on the existing django user model
@receiver(post_migrate)
def add_user_permissions(sender, **kwargs):
    from django.contrib.contenttypes.models import ContentType
    from django.contrib.auth.models import Permission
    try:
        content_type = ContentType.objects.get(app_label='auth', model='user')
    except ObjectDoesNotExist:
        return

    Permission.objects.get_or_create(codename='groupread', name='User may read documents of others in his/her group(s)', content_type=content_type)
    Permission.objects.get_or_create(codename='groupwrite', name='User may write documents of others in his/her group(s)', content_type=content_type)
    Permission.objects.get_or_create(codename='allowcopy', name='User may copy documents wherever he/she has write permission', content_type=content_type)
    Permission.objects.get_or_create(codename='allowdelete', name='User may delete documents wherever he/she has write permission', content_type=content_type)



import json
import importlib
from django.conf import settings

class Converter:
    def __init__(self, id, module, function, name, parameter_help="", parameter_default="", inputextensions=None): #pylint: disable=redefined-builtin
        self.id = id
        self.module = module
        self.function = function
        self.name = name
        self.parameter_help = parameter_help
        self.parameter_default = parameter_default
        if isinstance(inputextensions,str):
            self.inputextensions = [inputextensions]
        elif inputextensions:
            self.inputextensions = inputextensions
        else:
            self.inputextensions = []

    def get_output_name(self, inputfilename):
        for inputextension in self.inputextensions:
            if inputextension[0] == '.': inputextension = inputextension[1:]
            if inputfilename.lower().endswith('.' + inputextension.lower()):
                return inputfilename[:-len(inputextension)] + 'folia.xml'
        return inputfilename + '.folia.xml' #default, just append


    def parse_parameters(self, request, parameterfield, method='POST'):
        parameters =  json.loads('{' + getattr(request,method)[parameterfield] + '}')
        parameters['flatuser'] = request.user.username
        parameters['flatconfiguration'] =  settings.CONFIGURATIONS[request.session['configuration']]
        return parameters

    def convert(self, inputfilename, outputfilename, *args, **kwargs):
        try:
            module = importlib.import_module(self.module)
        except ImportError:
            return (False, "Module " + self.module + " could not be imported")
        try:
            convertfunction = getattr(module, self.function)
        except AttributeError:
            return (False, "Function " + self.function + " was not found in module " + self.module)
        response  = convertfunction(inputfilename, outputfilename, *args, **kwargs)
        if response is False:
            return (False, "Converter returned with an unspecified error")
        elif response is True:
            return True, "Conversion successful"
        elif len(response) == 2:
            return response
        else:
            return (False, "Converter did not return a compatible response, but be (bool succes, str errormessage)")


def get_converters(request):
    if 'convertors' in settings.CONFIGURATIONS[request.session['configuration']]:
        settings.CONFIGURATIONS[request.session['configuration']]['converters'] = settings.CONFIGURATIONS[request.session['configuration']]['convertors']
    if 'converters' in settings.CONFIGURATIONS[request.session['configuration']]:
        for converter in settings.CONFIGURATIONS[request.session['configuration']]['converters']:
            yield Converter(converter['id'] if 'id' in converter else converter['module'] +'.' + converter['function'], converter['module'], converter['function'], converter['name'], converter['parameter_help'] if 'parameter_help' in converter else "", converter['parameter_default'] if 'parameter_default' in converter else "", converter['inputextensions'] if 'inputextensions' in converter else [])


def inputformatchangefunction(request):
    def htmlescape(s):
        s = s.replace("'","&39;")
        s = s.replace('"',"&quot;")
        return s

    js = "if(this.value=='folia'){$('#convparameters').hide();$('#convparameterhelp').hide()};"
    for converter in get_converters(request):
        js += "if(this.value=='" + converter.id + "'){$('#convparameterhelp').html('" + htmlescape(converter.parameter_help) + "').show();$('#convparameters input').val('" + htmlescape(converter.parameter_default) + "');$('#convparameters').show()};"

    return js



#from django.conf.urls import patterns, include, url
from django import VERSION
if VERSION[0] > 1 or VERSION[1] >= 8: #Django 1.8 and higher
    from django.conf.urls import url, include
elif VERSION[1] >= 6: #Django 1.6
    from django.conf.urls import patterns, url, include
else:
    from django.conf.urls.defaults import *

import flat.modes.editor.views

urlpatterns = [
    # Examples:
    url(r'^pub/(?P<configuration>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/?$', flat.modes.editor.views.pub_view, name='view'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/history/?$', flat.modes.editor.views.history, name='history'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/revert/(?P<commithash>[a-f0-9]*)/?$', flat.modes.editor.views.revert, name='revert'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/save/?$', flat.modes.editor.views.save, name='save'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/?$', flat.modes.editor.views.view, name='view'),

    #redeclare viewer stuff: (why?)
    url(r'^pub/(?P<docid>[\w\d\-_\.]+)/poll/?$', flat.modes.viewer.views.pub_poll, name='poll'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/poll/?$', flat.modes.viewer.views.poll, name='poll'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/query/?$', flat.modes.viewer.views.query, name='query'),
    #url(r'^(?P<namespace>[\w\d\-_\./]+)/(?P<docid>[\w\d\-_\.]+)/(?P<elementid>[\w\d\-_\.]+)/?$', 'flat.modes.viewer.views.subview', name='subview'),
]

if VERSION[0] == 1 and VERSION[1] < 8: #Django <1.8
    urlpatterns = patterns('',*urlpatterns)



from django.db import models

class MetadataIndex(models.Model):
    namespace = models.CharField(max_length=255, db_index=True)
    docid = models.CharField(max_length=255, db_index=True)
    key = models.CharField(max_length=150)
    value = models.CharField(max_length=255)

    class Meta:
        verbose_name = "Metadata Index"
        verbose_name_plural = "Metadata Index"
        unique_together = (('namespace','docid','key'),)



from __future__ import print_function, unicode_literals, division, absolute_import
from django.shortcuts import render
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse, HttpResponseForbidden
from django.views.decorators.csrf import csrf_exempt
from django.conf import settings
import flat.comm
import flat.users
from flat.views import initdoc, query
import json
import sys
if sys.version < '3':
    from urllib2 import URLError
else:
    from urllib.error import URLError



@login_required
def view(request, namespace, docid):
    """The initial viewer, does not provide the document content yet"""
    return initdoc(request, namespace,docid, 'viewer', 'viewer.html')

@login_required
def poll(request, namespace, docid):
    if flat.users.models.hasreadpermission(request.user.username, namespace, request):
        try:
            r = flat.comm.get(request, '/poll/' + namespace + '/' + docid + '/', False)
        except URLError:
            return HttpResponseForbidden("Unable to connect to the document server [viewer/poll]")
        return HttpResponse(r, content_type='application/json')
    else:
        return HttpResponseForbidden("Permission denied")

@csrf_exempt
def pub_view(request, configuration, docid):
    """The initial viewer, does not provide the document content yet"""
    return initdoc(request, 'pub',docid, 'viewer', 'viewer.html', configuration=configuration)

@csrf_exempt
def pub_poll(request, docid):
    """The initial viewer, does not provide the document content yet"""
    try:
        r = flat.comm.get(request, '/poll/pub/' + docid + '/', False)
    except URLError:
        return HttpResponseForbidden("Unable to connect to the document server [viewer/poll]")
    return HttpResponse(r, content_type='application/json')



from __future__ import print_function, unicode_literals, division, absolute_import
from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse,HttpResponseForbidden
from django.conf import settings
import flat.comm
import flat.users
from flat.views import getcontext
import json


@login_required
def view(request, namespace, docid):
    if flat.users.models.hasreadpermission(request.user.username, namespace, request):
        try:
            doc = flat.comm.query(request, "USE " + namespace + "/" + docid + " SELECT ALL FORMAT flat", setdefinitions=True,declarations=True) #get the entire document with meta information
        except URLError:
            return HttpResponseForbidden("Unable to connect to the document server [viewer/view]")
        d = getcontext(request,namespace,docid, doc, 'structureeditor')
        return render(request, 'structureeditor.html', d)
    else:
        return HttpResponseForbidden("Permission denied")


#@login_required
#def annotate(request,namespace, docid):
#    if flat.users.models.haswritepermission(request.user.username, namespace, request):
#        d = flat.comm.postjson(request, '/annotate/' +namespace + '/' + docid + '/', request.body)
#        return HttpResponse(json.dumps(d), mimetype='application/json')
#    else:
#        return HttpResponseForbidden("Permission denied, no write access")







#! /usr/bin/env python3
# -*- coding: utf8 -*-

from __future__ import print_function

import os
import sys
from setuptools import setup


try:
   os.chdir(os.path.dirname(sys.argv[0]))
except:
   pass


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

if sys.version >= '3':
    extradeps = ['foliadocserve >= 0.6'] #change also in comm.py!
else:
    extradeps = []

setup(
    name = "FoLiA-Linguistic-Annotation-Tool",
    version = "0.7.5", #Also change in flat/__init__.py !!
    author = "Maarten van Gompel",
    author_email = "proycon@anaproy.nl",
    description = ("FLAT is a web-based linguistic annotation environment based around the FoLiA format (https://proycon.github.io/folia), a rich XML-based format for linguistic annotation. Flat allows users to view annotated FoLiA documents and enrich these documents with new annotations, a wide variety of linguistic annotation types is supported through the FoLiA paradigm."),
    license = "GPL",
    keywords = "flat linguistic annotation nlp computational_linguistics folia annotator web",
    url = "https://github.com/proycon/flat",
    packages=['flat','flat.modes','flat.modes.structureeditor','flat.modes.viewer','flat.modes.editor','flat.modes.metadata','flat.users'],
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Text Processing :: Linguistic",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Operating System :: POSIX",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    zip_safe=False,
    package_data = {'flat':['templates/*.html','style/*'], 'flat.modes.structureeditor':['templates/*.html'],  'flat.modes.viewer':['templates/*.html'], 'flat.modes.editor':['templates/*.html'], 'flat.modes.metadata':['templates/*.html']  },
    install_requires=['pynlpl >= 1.1.2','Django >= 1.8','requests'] + extradeps
)



#from django.conf.urls import patterns, include, url
from django import VERSION
if VERSION[0] > 1 or VERSION[1] >= 8: #Django 1.8 and higher
    from django.conf.urls import url, include
elif VERSION[1] >= 6: #Django 1.6
    from django.conf.urls import patterns, url, include
else:
    from django.conf.urls.defaults import *

import flat.modes.viewer.views

urlpatterns = [
    # Examples:
    url(r'^pub/(?P<docid>[\w\d\-_\.]+)/poll/?$', flat.modes.viewer.views.pub_poll, name='poll'),
    url(r'^pub/(?P<configuration>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/?$', flat.modes.viewer.views.pub_view, name='view'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/poll/?$', flat.modes.viewer.views.poll, name='poll'),
    url(r'^(?P<namespace>[\w\d\-_\./@]+)/(?P<docid>[\w\d\-_\.]+)/?$', flat.modes.viewer.views.view, name='view'),
]

if VERSION[0] == 1 and VERSION[1] < 8: #Django <1.8
    urlpatterns = patterns('',*urlpatterns)



from __future__ import print_function, unicode_literals, division, absolute_import
from django.shortcuts import render, redirect
from django.contrib.auth.decorators import login_required
from django.http import HttpResponse,HttpResponseForbidden
import sys
if sys.version < '3':
    from urllib2 import URLError
else:
    from urllib.error import URLError
from django.views.decorators.csrf import csrf_exempt
from django.conf import settings
import flat.comm
import flat.users
from flat.views import initdoc, fatalerror
import json



@login_required
def view(request, namespace, docid):
    """The initial view, does not provide the document content yet"""
    if flat.users.models.hasreadpermission(request.user.username, namespace, request):
        if 'autodeclare' in settings.CONFIGURATIONS[request.session['configuration']]:
            if flat.users.models.haswritepermission(request.user.username, namespace, request):
                for annotationtype, set in settings.CONFIGURATIONS[request.session['configuration']]['autodeclare']:
                    try:
                        r = flat.comm.query(request, "USE " + namespace + "/" + docid + " DECLARE " + annotationtype + " OF " + set)
                    except Exception as e:
                        return fatalerror(request,e)

        return initdoc(request, namespace,docid, 'editor', 'editor.html')
    else:
        return fatalerror(request,"Permission denied")


@csrf_exempt
def pub_view(request, docid, configuration):
    """The initial view, does not provide the document content yet"""
    if 'autodeclare' in settings.CONFIGURATIONS[configuration]:
        for annotationtype, set in settings.CONFIGURATIONS['configuration']['autodeclare']:
            try:
                r = flat.comm.query(request, "USE pub/" + docid + " DECLARE " + annotationtype + " OF " + set)
            except Exception as e:
                return fatalerror(request,e)

    return initdoc(request, 'pub',docid, 'editor', 'editor.html', configuration=configuration)

@login_required
def history(request,namespace, docid):
    if namespace == 'pub' or flat.users.models.hasreadpermission(request.user.username, namespace, request):
        try:
            if hasattr(request, 'body'):
                d = flat.comm.get(request, '/getdochistory/' +namespace + '/' + docid + '/',False)
            else:
                d = flat.comm.get(request, '/getdochistory/' +namespace + '/' + docid + '/',False)
        except URLError as e:
            return HttpResponseForbidden("Unable to connect to the document server: " + e.reason + " [editor/history]")
        return HttpResponse(d, content_type='application/json')
    else:
        return HttpResponseForbidden("Permission denied, no read access")

@login_required
def revert(request,namespace, docid, commithash):
    if namespace == 'pub' or flat.users.models.haswritepermission(request.user.username, namespace, request):
        try:
            if hasattr(request, 'body'):
                flat.comm.get(request, '/revert/' +namespace + '/' + docid + '/?commithash=' + commithash,False)
            else:
                flat.comm.get(request, '/revert/' +namespace + '/' + docid + '/?commithash=' + commithash,False)
        except URLError as e:
            return HttpResponseForbidden("Unable to connect to the document server: " + e.reason + " [editor/revert]")
        return HttpResponse("{}", content_type='application/json') #no content, client will do a full reload
    else:
        return HttpResponseForbidden("Permission denied, no write access")

@login_required
def save(request,namespace, docid):
    if namespace == 'pub' or flat.users.models.haswritepermission(request.user.username, namespace, request):
        if 'message' in request.GET:
            message = request.GET['message']
        else:
            message = ""
        try:
            if hasattr(request, 'body'):
                flat.comm.get(request, '/save/' +namespace + '/' + docid + '/?message=' + message,False)
            else:
                flat.comm.get(request, '/save/' +namespace + '/' + docid + '/?message=' + message,False)
        except URLError as e:
            return HttpResponseForbidden("Unable to connect to the document server: " + e.reason + " [editor/revert]")
        return HttpResponse("{}", content_type='application/json')
    else:
        return HttpResponseForbidden("Permission denied, no write access")
